package cz.mendelu.dnaAnalyser.analyse.cpg;


import cz.mendelu.dnaAnalyser.batch.BatchService;
import cz.mendelu.dnaAnalyser.sequence.Sequence;
import cz.mendelu.dnaAnalyser.sequence.data.SequenceData;
import cz.mendelu.dnaAnalyser.sequence.data.SequenceDataRepository;
import cz.mendelu.dnaAnalyser.utils.analyse.service.heatmap.Heatmap;
import cz.mendelu.dnaAnalyser.utils.exception.NotFoundException;
import cz.mendelu.dnaAnalyser.utils.repository.BaseRepository;
import cz.mendelu.dnaAnalyser.utils.service.ModelService;
import lombok.extern.slf4j.Slf4j;
import org.apache.ibatis.session.RowBounds;
import org.apache.ibatis.session.SqlSession;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.Pageable;
import org.springframework.data.domain.PageImpl;
import org.springframework.data.domain.Sort;

import java.util.*;

@Service
@Slf4j
public class CpgService extends ModelService<Cpg> {

    @Autowired
    private CpgRepository cpgRepository;

    @Autowired
    private SequenceDataRepository sequenceDataRepository;

    @Autowired
    private BatchService batchService;

    @Autowired
    private CpgAnalyser cpgAnalyser;

    @Autowired
    private CpgDataSessionService cpgDataSessionService;

    @Override
    protected BaseRepository<Cpg, UUID> getRepository() {
        return cpgRepository;
    }

    public Cpg executeCpgAnalysis(Cpg cpg) {

        cpg = cpgRepository.save(cpg);
        batchService.createBatch(cpg.getOwner(), cpg)
                .execute(uuid -> {
                    Cpg analysis = cpgRepository
                            .findById(uuid)
                            .orElseThrow(() -> new NotFoundException(Cpg.class, uuid));
                    Sequence sequence = analysis.getSequence();
                    SequenceData sequenceData = sequenceDataRepository.load(sequence);

                    int minWindowSize = analysis.getMinWindowSize();
                    double minGcPercentage = analysis.getMinGcPercentage();
                    double minObservedToExpectedCpG = analysis.getMinObservedToExpectedCpG();
                    int minIslandMergeGap = analysis.getMinIslandMergeGap();
                    char firstNucleotide = analysis.getFirstNucleotide();
                    char secondNucleotide = analysis.getSecondNucleotide();

                    if (sequence.getLength() < minWindowSize) {
                        throw new IllegalArgumentException(String.format("The window (%d) must be smaller than the sequence length (%d).", minWindowSize, sequence.getLength()));
                    }

                    try (SqlSession sqlSession = cpgDataSessionService.openDataSession(analysis)) {
                        CpgMapper cpgMapper = sqlSession.getMapper(CpgMapper.class);

                        ArrayList<CpgAnalyseResult> results = cpgAnalyser.getResults(sequenceData.toPlain(), minWindowSize, minGcPercentage, minObservedToExpectedCpG, minIslandMergeGap, firstNucleotide, secondNucleotide);

                        results.forEach(cpgMapper::insert);

                        analysis.setResultCount((long) results.size());
                        analysis.finishNow();
                        sqlSession.commit();
                    }
                    cpgRepository.save(analysis);
                }).start();
        return cpg;
    }

    public Page<CpgAnalyseResult> returnCpgAnalysis(UUID id, Integer sequenceStart, Integer sequenceLength, Pageable pageable) {
        Cpg analysis = findOne(id);

        int start = (sequenceStart == null) ? 0 : sequenceStart;
        int end = start + ((sequenceLength == null) ? analysis.getSequence().getLength() : sequenceLength);

        try (SqlSession sqlSession = cpgDataSessionService.openDataSession(analysis)) {
            CpgMapper cpgMapper = sqlSession.getMapper(CpgMapper.class);

            RowBounds rowBounds = new RowBounds((int) pageable.getOffset(), pageable.getPageSize());
            String orderName = "POSITION";
            Sort.Direction orderDirection = Sort.Direction.ASC;
            Sort.Order order = getSortOrderOrNull(pageable);
            if (order != null) {
                orderName = order.getProperty().toUpperCase();
                orderDirection = order.getDirection();
            }

            List<CpgAnalyseResult> list;
            if (orderDirection == Sort.Direction.ASC) {
                list = cpgMapper.getAllBetweenPositionOrderAsc(start, end, orderName, rowBounds);
            } else {
                list = cpgMapper.getAllBetweenPositionOrderDesc(start, end, orderName, rowBounds);
            }
            int total = cpgMapper.countAllBetweenPosition(start, end);

            return new PageImpl<>(list, pageable, total);
        }
    }


    public Heatmap getHeatmap(UUID id, Integer segmentsCount, Integer from, Integer to) {
        Cpg analysis = findOne(id);
        return cpgDataSessionService.heatmap(CpgMapper.class, analysis, segmentsCount, from, to);
    }

    public Cpg modifyTags(UUID id, Set<String> tags) {
        Cpg analysis = cpgRepository.findById(id)
                .orElseThrow(() -> new NotFoundException(Cpg.class, id));
        analysis.setTags(tags);
        cpgRepository.save(analysis);
        return analysis;
    }

    public int avgCpgLen(UUID id) {
        Cpg analyse = findOne(id);
        int result;

        try (SqlSession sqlSession = cpgDataSessionService.openDataSession(analyse)) {
            CpgMapper cpgMapper = sqlSession.getMapper(CpgMapper.class);
            ArrayList<CpgAnalyseResult> analysisCpg = cpgMapper.getAll();

            int totalCpgLen = analysisCpg.stream()
                    .mapToInt(e -> e.length)
                    .sum();

            result = !analysisCpg.isEmpty()
                    ? totalCpgLen / analysisCpg.size()
                    : 0;
        }

        return result;
    }

    private Sort.Order getSortOrderOrNull(Pageable pageable) {
        if (pageable == null) return null;
        Iterator<Sort.Order> orderIterator = pageable.getSort().iterator();
        if (!orderIterator.hasNext()) return null;
        return orderIterator.next();
    }

    public Cpg findOne(UUID id) {
    return cpgRepository.findById(id)
            .orElseThrow(() -> new NotFoundException(Cpg.class, id));
}
}
