package cz.mendelu.dnaAnalyser.analyse.cpg;

import java.util.regex.Pattern;

public class CpgWindow {

    private final Pattern patternGorC;
    private final int MIN_WINDOW_SIZE;
    private final char SECOND_NUCLEOTIDE;
    private final char FIRST_NUCLEOTIDE;
    private final double MIN_OBSERVED_TO_EXPECTED_CPG;
    private final double MIN_GC_PERCENTAGE;
    public String record;
    private final int record_len;
    public int window_begin;
    public int window_size;
    public int window_end;
    private long gc_count;
    private int obs_cpg;
    private double gc_perc;
    private double obs_exp;



    public CpgWindow(String record, int window_begin, int window_size, int MIN_WINDOW_SIZE, double MIN_GC_PERCENTAGE, double MIN_OBSERVED_TO_EXPECTED_CPG, char FIRST_NUCLEOTIDE, char SECOND_NUCLEOTIDE) {
        this.record = record;
        this.patternGorC = Pattern.compile("(?=(" + FIRST_NUCLEOTIDE + "|" + SECOND_NUCLEOTIDE + "))");
        this.record_len = record.length();
        this.window_begin = window_begin;
        this.window_size = window_size;
        this.window_end = window_begin + window_size;
        this.gc_count = 0;
        this.obs_cpg = 0;
        this.gc_perc = 0;
        this.obs_exp = 0;
        this.MIN_GC_PERCENTAGE = MIN_GC_PERCENTAGE;
        this.MIN_OBSERVED_TO_EXPECTED_CPG = MIN_OBSERVED_TO_EXPECTED_CPG;
        this.FIRST_NUCLEOTIDE = FIRST_NUCLEOTIDE;
        this.SECOND_NUCLEOTIDE = SECOND_NUCLEOTIDE;
        this.MIN_WINDOW_SIZE = MIN_WINDOW_SIZE;
        update_gc_count();
    }

    public String toString() {
        return "Window at (" + window_begin + ", " + window_end + ")\tgcper:" + gc_perc + "\tobs_exp:" + obs_exp + "\twin_length:" + window_size;
    }

    public boolean is_island() {
        _evaluate();
        return gc_perc > MIN_GC_PERCENTAGE && obs_exp > MIN_OBSERVED_TO_EXPECTED_CPG;
    }

    private void _evaluate() {
        gc_perc = gc_count / (double) window_size;
        double exp_cpg = Math.pow((double) gc_count / 2, 2) / window_size;
        try {
            obs_exp = obs_cpg / exp_cpg;
        } catch (ArithmeticException e) {
            obs_exp = 0.0;
        }
    }

    public void update_gc_count() {
        if (this.window_end > record_len) {
            this.window_end = record_len;
        }
        String record_cut = record.substring(window_begin, window_end);
        gc_count = this.patternGorC.matcher(record_cut).results().count();
        obs_cpg = countOccurrences(record_cut, FIRST_NUCLEOTIDE + String.valueOf(SECOND_NUCLEOTIDE));
        _evaluate();
    }

    public void shrink_left() {
        if (charAt(record, window_begin) == FIRST_NUCLEOTIDE || charAt(record, window_begin) == SECOND_NUCLEOTIDE) {
            gc_count -= 1;
        }
        if (charAt(record, window_begin) == FIRST_NUCLEOTIDE && charAt(record,window_begin + 1) == SECOND_NUCLEOTIDE) {
            obs_cpg -= 1;
        }
        window_begin += 1;
    }

    public boolean expand_right() {
        if (window_end >= record_len) {
            return false;
        }
        if (charAt(record, window_end) == FIRST_NUCLEOTIDE || charAt(record, window_end) == SECOND_NUCLEOTIDE) {
            gc_count += 1;
        }
        if (charAt(record,window_end - 1) == FIRST_NUCLEOTIDE && charAt(record,window_end) == SECOND_NUCLEOTIDE) {
            obs_cpg += 1;
        }
        window_end += 1;
        return true;
    }

    public void shrink_right() {
        if (charAt(record,window_end - 1) == FIRST_NUCLEOTIDE || charAt(record, window_end - 1) == SECOND_NUCLEOTIDE) {
            gc_count -= 1;
        }
        if (charAt(record, window_end - 2) == FIRST_NUCLEOTIDE && charAt(record,window_end - 1) == SECOND_NUCLEOTIDE) {
            obs_cpg -= 1;
        }
        window_end -= 1;
    }

    private void expand_left() {
        if (charAt(record,window_begin - 1) == FIRST_NUCLEOTIDE || charAt(record,window_begin - 1) == SECOND_NUCLEOTIDE) {
            gc_count += 1;
        }
        if (charAt(record, window_begin - 1) == FIRST_NUCLEOTIDE && charAt(record, window_begin) == SECOND_NUCLEOTIDE) {
            obs_cpg += 1;
        }
        window_begin -= 1;
    }

    public boolean expand_right_island() {
        int remaining_window_length = record_len - window_end;
        if (remaining_window_length == 0) {
            return false;
        }
        int new_window_begin;
        if (remaining_window_length < MIN_WINDOW_SIZE) {
            System.out.println("reached end of sequence in big jump, be careful");
            new_window_begin = record_len - MIN_WINDOW_SIZE;
        } else {
            new_window_begin = window_end;
        }
        CpgWindow next_window = new CpgWindow(record, new_window_begin, MIN_WINDOW_SIZE, MIN_WINDOW_SIZE, MIN_GC_PERCENTAGE, MIN_OBSERVED_TO_EXPECTED_CPG, FIRST_NUCLEOTIDE, SECOND_NUCLEOTIDE);
        window_end = new_window_begin + MIN_WINDOW_SIZE;
        window_size = window_end - window_begin;
        return next_window.is_island();
    }

    public void rollback_until_island() {
        CpgWindow last_window = new CpgWindow(record, window_end - MIN_WINDOW_SIZE, MIN_WINDOW_SIZE, MIN_WINDOW_SIZE, MIN_GC_PERCENTAGE, MIN_OBSERVED_TO_EXPECTED_CPG, FIRST_NUCLEOTIDE, SECOND_NUCLEOTIDE);
        while (!last_window.is_island()) {
            last_window.expand_left();
            last_window.shrink_right();
        }
        join_window(last_window);
    }

    public void join_window(CpgWindow other_window) {
        if (other_window.window_end > this.window_begin) {
            window_end = other_window.window_end;
            window_size = window_end - window_begin;
            update_gc_count();
        }
    }

    private static int countOccurrences(String str, String sub) {
        int count = 0;
        int idx = 0;
        while ((idx = str.indexOf(sub, idx)) != -1) {
            count++;
            idx += sub.length();
        }
        return count;
    }

    public static char charAt(String str, int index) {
        if (str == null) {
            throw new NullPointerException("String is null");
        }

        // If the index is negative, adjust it to access the character from the end of the string
        if (index < 0) {
            index += str.length();
        }

        // Check if the adjusted index is still out of range
        if (index < 0 || index >= str.length()) {
            throw new StringIndexOutOfBoundsException("String index out of range: " + index);
        }

        return str.charAt(index);
    }

    public double getObsExp() {
        return obs_exp;
    }

    public double getGcPerc() {
        return gc_perc;
    }
}
