package cz.mendelu.dnaAnalyser.analyse.rloopr;

import cz.mendelu.dnaAnalyser.sequence.stream.BufferedWindow;
import cz.mendelu.dnaAnalyser.sequence.stream.Window;

import org.springframework.stereotype.Service;
import lombok.extern.slf4j.Slf4j;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.LinkedList;
import java.util.List;
import java.util.Queue;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.stream.Collectors;
import java.util.stream.IntStream;
import java.util.stream.Stream;

/**
 * This Service class provide algorithms for rloop analysis.
 */
@Slf4j
@Service
public class RloopR {

    /**
     * FInd R-loop Initalization zone for a given window
     *
     * @param window with sequence with min length 20 nuclide.
     * @param complement finds all R-loops for complementary strand
     * @return singleResult with Data.
     * @throws IllegalArgumentException if sequence has wrong length.
     */
    public ArrayList<RlooprResult> getInitiationZones(short modelRIZ, Window window, boolean complement) {

        ArrayList<RlooprResult> allResults = new ArrayList<RlooprResult>();

        RlooprResult singleResult;

        Window windowCopy = window.getSubWindow(0, window.toPlain().length());
        Window reverseComplement = BufferedWindow.wrap(getComplement(windowCopy.toPlain()));

        String dataPlain = complement ? reverseComplement.toPlain() : window.toPlain();

        Integer maxOffset = 2000;

        Boolean additionalMetrics = true; // addition metrics which may slow the computation for large sequences

        /**
         * Regex pattern explanation:
         * 
         * (G{3,}) - group of at least 3 guanines in a row ([ACGT]{1,10}?) - any
         * nucleotide in the range (1, 10) with non-greedy search Non greedy search is
         * used, becase without it, it would take the last Guanine
         * 
         * e.g. groups for sequence with greedy search GGG|A|GGG|CTTG|GGG e.g. groups
         * for sequence without greedy search GGG|A|GGG|CTT|GGGG <- !
        */
        
        final Pattern model1Pattern = Pattern.compile("(G{3,})([ACGT]{1,10}?)(G{3,})([ACGT]{1,10}?)(G{3,})");
        final Pattern model2Pattern = Pattern.compile("(G{4,})([ACGT]{1,10}?)(G{4,})");


        Matcher match;

        switch (modelRIZ) {
            case 0:
                match = model1Pattern.matcher(dataPlain);
                break;
            case 1:
                match = model2Pattern.matcher(dataPlain);
                break;
            default:
                match = model1Pattern.matcher(dataPlain);
        }

        Queue<Window> windowQueue = new LinkedList<>();
        while (match.find()) { // add match start to queue so we can divide and conquer them all at once
            if (complement) {
                windowQueue.add(reverseComplement.getSubWindow(match.start(), match.end()+maxOffset));
            }
            else {
                windowQueue.add(window.getSubWindow(match.start(), match.end()+maxOffset));
            }
        }
        match.reset();

        // this part can be nicely parelelized, because we know the window size is maximum of 2050 bases
        // and we have start of each match

        while (match.find()) {
            singleResult = new RlooprResult(); // new data object for each match
            String riz = match.group();
            Integer rizLen = riz.length();

            Window currentWindow = windowQueue.remove();
            Integer currentWindowSize = currentWindow.getSize();
            String windowPlainData = currentWindow.toPlain();

            Integer lenWithoutRiz = currentWindowSize - rizLen;
            if (lenWithoutRiz < 100 || !checkGRichness(riz, 0.5)) { // either the string is too short or there isnt enough guanine
                continue; // invalid data for the analysis
            }

            int maxRezIter = (lenWithoutRiz < maxOffset) ? lenWithoutRiz : maxOffset; // analyse max 2050 nocleotides, or take the shorter input as a whole

            String rezCheckArea = windowPlainData.substring(rizLen, windowPlainData.length());
            boolean rloopFormed = false;
            

            int linkerIt = 0;
            int rezIt = 0;
            Stream.Builder<String> builder = Stream.builder(); 

            for (linkerIt=0; linkerIt < 50; linkerIt++) {
                for (rezIt=linkerIt + 100; rezIt < (maxOffset - rezIt); rezIt++) {               
                    try {
                        builder.add(rezCheckArea.substring(linkerIt, rezIt));
                    } catch (StringIndexOutOfBoundsException e) {
                        // last subwindow causes trouble
                        break;
                    }
                }
            }

            Stream<String> linkerList = builder.build();
            List<String> elongedRezList = linkerList
                                                .parallel()
                                                .filter(seq -> checkGRichness(seq, 0.4))
                                                .collect(Collectors.toList());

            /*
            * Writing down the results
            */                                    
            if (elongedRezList.size() > 0) {
                String finalRez = Collections.max(elongedRezList, Comparator.comparing(String::length));
                int linkerLen = rezCheckArea.indexOf(finalRez);

                String modelStr;
                switch(modelRIZ) {
                    case 0:
                        modelStr = "RIZ 3G cluster (m1)";
                        break;
                    case 1:
                        modelStr = "RIZ 4G cluster (m2)";
                        break;
                    default:
                        modelStr = "";
                        break;
                }

                singleResult.rez = finalRez;
                singleResult.riz = riz;
                singleResult.linker = rezCheckArea.substring(0, linkerLen);
                singleResult.model = modelStr;
                singleResult.strand = complement ? '-' : '+';
                singleResult.length = riz.length()+linkerLen+singleResult.rez.length();
                singleResult.position = complement ? dataPlain.length()-match.start()-singleResult.length : match.start();
                singleResult.id = null;
                rloopFormed = true;
                
                // additional metrics
                String rloop = singleResult.riz + singleResult.linker + singleResult.rez;

                if (additionalMetrics) { // additional metrics, which may slow computation for large sequences
                    singleResult.rizGRichness = gRichness(riz);
                    singleResult.rloopGRichness = gRichness(rloop);
    
                    singleResult.g3cnt = gClusters(rloop, 3);
                    singleResult.g4cnt = gClusters(rloop, 4);
                    singleResult.gNcnt = gClusters(rloop, 5);
                }

            }

            if (singleResult.rez != null && rloopFormed) {
                allResults.add(singleResult);
            }
        }

        return allResults;
    }

    /**
     * Checks nucleotide sequence if it is Guanine rich enough
     *
     * @param sequence         inspected sequence
     * @param amount           comparison threshold (usually 0.4, 0.5)
     * @return boolean value if it fits the condition
     */

    protected Boolean checkGRichness(String sequence, double amount) {
        return (gRichness(sequence) >= amount);
    }

    /**
     * Returns sequence G richness
     *
     * @param sequence         inspected sequence
     * @return double richness
     */

    protected double gRichness(String sequence) {
        long gCount = sequence.chars().filter(ch -> ch == 'G').count();
        return (double) gCount / (double) sequence.length();
    }

     /**
     * Returns number of clusters given their size
     *
     * @param clusterSize size of G cluster
     * @param rloop riz + linker + rez
     * @return long clusters
     */

    protected long gClusters(String rloop, int clusterSize) {

        final Pattern clusterPattern = Pattern.compile(String.format("G{%d,}", clusterSize));
        Matcher match = clusterPattern.matcher(rloop);

        int matchCnt = 0;
        while (match.find()) {
            matchCnt++;
        }

        return matchCnt;
    }

    protected String getComplement(String strand) {
        // temporary function to return the complementary strand for DNA
        // replaces Cytosine for Guanine and backwards

        StringBuilder reverse = new StringBuilder(strand).reverse();

        return reverse.toString()
                .replace('G', '1')
                .replace('C', 'G')
                .replace('1', 'C')
                // G -C swapped
                .replace('A', '2')
                .replace('T', 'A')
                .replace('2', 'T');
    }

}
