package cz.mendelu.dnaAnalyser.overlay;

import cz.mendelu.dnaAnalyser.analyse.g4hunter.G4Hunter;
import cz.mendelu.dnaAnalyser.analyse.g4hunter.G4HunterService;
import cz.mendelu.dnaAnalyser.analyse.g4hunter.quadruplex.*;
import cz.mendelu.dnaAnalyser.analyse.rloopr.*;
import cz.mendelu.dnaAnalyser.sequence.annotation.Annotation;
import cz.mendelu.dnaAnalyser.sequence.annotation.AnnotationService;
import cz.mendelu.dnaAnalyser.sequence.annotation.feature.Feature;
import cz.mendelu.dnaAnalyser.utils.analyse.model.Analyse;
import cz.mendelu.dnaAnalyser.utils.analyse.repository.AnalyseRepository;
import lombok.AllArgsConstructor;
import lombok.Data;
import org.apache.ibatis.session.SqlSession;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import java.lang.reflect.Field;
import java.util.*;

@Service
public class OverlayService {

    @Autowired
    private AnnotationService annotationService;

    @Autowired
    private G4HunterService g4HunterService;

    @Autowired
    private RlooprService rlooprService;

    @Autowired
    public AnalyseRepository analyseRepository;

    @Autowired
    public QuadruplexDataSessionService quadruplexDataSessionService;

    @Autowired
    private RloopDataSessionService rloopDataSessionService;

    @Data
    @AllArgsConstructor
    static class Group {

        public Integer[] subGroup;

        public Group(Group group) {
            this.subGroup = group.subGroup;
        }
    }

    @Data
    @AllArgsConstructor
    static class Result {

        public String label;

        public Group[] group;

    }


    public Map<String, OverlayItem> annotationOverlay(List<OverlayController.SourceItem> sources, Integer areaSize){
        if(areaSize < 0 || areaSize > 1000){
            throw new IllegalArgumentException("Area size must be within (0,1000) interval");
        }
        List<Feature> annotationList = new ArrayList<>();
        List<Feature> annotations = new ArrayList<>();
        Map<String, OverlayItem> result = new HashMap<>();
        UUID analyseID = null;
        List<Double> intervals = null;
        String groupBy = null;
        List<String> labels = null;
        Boolean aggregated = false;
        List<String> groupIntervals = new ArrayList<>();
        SqlSession session = null;

        for (OverlayController.SourceItem element : sources) {
            if (element.getType().equals("annotation")){
                UUID annotationID = UUID.fromString(element.getId().toString());
                Annotation annotation = annotationService.findOne(annotationID);
                annotations = annotationService.getFeatures(annotation);
                annotationList = createAnnotationList(annotations, areaSize);
            }else{
                analyseID = element.getId();
                intervals = element.getIntervals();
                groupBy = element.getGroupBy();
                aggregated = element.getAggregated();
            }

        }
        OverlayController.SourceItem firstItem = sources.get(0);
        if(firstItem.getType().equals("annotation")){
            if(firstItem.getGroupBy().equals("featureType")){
                labels = getLabels(annotations);
            }
        }
        String labelGroupBy = sources.get(0).getGroupBy();

        if(intervals != null){
            groupIntervals = createIntervals(intervals);
            for (String label : labels){
                OverlayItem overlay = new OverlayItem();
                Map<String, Object> group = new HashMap<>();

                groupIntervals.forEach(item ->{
                    EnumMap<Occurence, Integer> occurenceMap = new EnumMap<>(Occurence.class);
                    occurenceMap.put(Occurence.BEFORE, 0);
                    occurenceMap.put(Occurence.IN, 0);
                    occurenceMap.put(Occurence.AFTER, 0);
                    group.put(item, occurenceMap);
                });

                overlay.setOccureneces(group);
                overlay.setAverageCoverage(0.0);
                overlay.setAverageLength(0.0);
                overlay.setTotalLength(0.0);
                result.put(label,overlay);
            }
        }else{
            if(labels != null){
                for (String label : labels){
                    OverlayItem overlay = new OverlayItem();
                    Map<String, Object> group = new HashMap<>();
                    overlay.setOccureneces(group);
                    overlay.setAverageCoverage(0.0);
                    overlay.setAverageLength(0.0);
                    overlay.setTotalLength(0.0);
                    result.put(label,overlay);
                }
            }
        }

        Analyse analyse = analyseRepository.findOne(analyseID);
        if (analyse instanceof G4Hunter) {
            session = quadruplexDataSessionService.openDataSession(analyse);
        }else if (analyse instanceof Rloop){
            session = rloopDataSessionService.openDataSession(analyse);
        }

        for (int i = 0; i <= annotationList.size()-1; i ++){
        //for (Feature annotation : annotationList){
            Feature annotationFeature = annotationList.get(i);
            int from = Integer.parseInt(annotationFeature.getStartPosition());
            int to = Integer.parseInt(annotationFeature.getEndPosition());
            result = createOverlay(analyse, annotationFeature, session, result, groupBy, labelGroupBy, intervals, groupIntervals, aggregated, from-areaSize, to+areaSize);
        }
        return result;


    }


    public Map<String, OverlayItem> createOverlay(Analyse analyse, Feature annotation, SqlSession session, Map<String, OverlayItem> result, String groupBy, String labelGroupBy, List<Double> intervals, List<String> groupIntervals, Boolean aggregated, Integer from, Integer to) {
        assert analyse.getSequence() != null && analyse.getSequence().getLength() > 0 : "Analyse has no sequence or sequence length is 0";
        if (analyse instanceof G4Hunter){
            if(aggregated){
                result = quadruplexDataSessionService.annotationOverlay(AggregateQuadruplexMapper.class, annotation, session, result, groupBy, labelGroupBy, intervals, groupIntervals, from, to);
            }else{
                result = quadruplexDataSessionService.annotationOverlay(QuadruplexMapper.class, annotation, session, result, groupBy, labelGroupBy, intervals, groupIntervals, from, to);
            }
        }else if (analyse instanceof Rloop){
            result = rloopDataSessionService.annotationOverlay(RloopMapper.class, annotation, session, result, groupBy, labelGroupBy, intervals, groupIntervals, from, to);
        }

        return result;

    }

    private List<String> createIntervals(List<Double> intervals){
        List<String> output = new ArrayList<>();
        for (int i=0; i < intervals.size()-1; i++){
            output.add(intervals.get(i)+"-"+intervals.get(i+1));
        }
        return output;
    }

    private List<String> getLabels(List<Feature> features){
        List<String> output = new ArrayList<>();
        features.forEach(feature -> {
            if(!output.contains(feature.getFeatureType())){
                output.add(feature.getFeatureType());
            }
        });
        return output;
    }


    private List<Feature> createAnnotationList(List<Feature> features, Integer area){
        List<Feature> output = new ArrayList<>();
        features.forEach(feature -> {
            output.add(new Feature(feature.getFeatureType(), feature.getStartPosition(), feature.getEndPosition(), area));
        });
        return output;
    }


}
