package cz.mendelu.dnaAnalyser.sequence;

import cz.mendelu.dnaAnalyser.jwt.JwtAuthenticationUtils;
import cz.mendelu.dnaAnalyser.sequence.inport.ImportSequenceFormat;
import cz.mendelu.dnaAnalyser.sequence.inport.ImportSequenceFrom;
import cz.mendelu.dnaAnalyser.sequence.inport.file.ImportSequenceFromFileRequest;
import cz.mendelu.dnaAnalyser.sequence.inport.text.ImportSequenceFromTextRequest;
import cz.mendelu.dnaAnalyser.user.User;
import cz.mendelu.dnaAnalyser.utils.controller.DeleteMapping;
import cz.mendelu.dnaAnalyser.utils.controller.GetMapping;
import cz.mendelu.dnaAnalyser.utils.controller.PostMapping;
import cz.mendelu.dnaAnalyser.utils.controller.PutMapping;
import cz.mendelu.dnaAnalyser.utils.controller.*;
import cz.mendelu.dnaAnalyser.utils.exception.InternalServerException;
import cz.mendelu.dnaAnalyser.utils.model.TagUtils;
import cz.mendelu.dnaAnalyser.utils.service.JsonService;
import cz.mendelu.dnaAnalyser.utils.swagger.ApiPaginationParams;
import io.swagger.annotations.Api;
import io.swagger.annotations.ApiOperation;
import lombok.Data;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.Pageable;
import org.springframework.http.HttpStatus;
import org.springframework.http.MediaType;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.security.core.Authentication;
import org.springframework.web.bind.annotation.*;
import org.springframework.web.bind.annotation.PatchMapping;
import org.springframework.web.multipart.MultipartFile;
import springfox.documentation.annotations.ApiIgnore;

import javax.validation.Valid;
import javax.validation.constraints.NotNull;
import java.nio.charset.StandardCharsets;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.UUID;
import java.util.stream.Collectors;


/**
 * Created by Jan Kolomazník on 24.3.17.
 */
@RestController
@Api(tags = "api/sequence")
@RequestMapping("api/sequence")
public class SequenceController {

    @Data
    static public class PatchSequenceRequest {

        private boolean length;
        private boolean nucleicCounts;
    }

    @Autowired
    private SequenceService sequenceService;

    @Autowired
    private JsonService jsonService;

    @GetMapping
    @ResponseStatus(HttpStatus.OK)
    @PreAuthorize("isAuthenticated()")
    @ApiPaginationParams
    @ApiOperation("Get page with sequences")
    public ResponsePage<Sequence> getPageWithSequences(
            PaginationParam paginationParam,
            SortParam sortParam,
            TagFilterParam tagFilterParam
    ) {
        Pageable pageable = paginationParam.toPageable(sortParam.toSort());
        Page<Sequence> sequencePage = sequenceService.findAll(pageable, tagFilterParam);
        return ResponsePage.responsePage(sequencePage);
    }

    @GetMapping("/tag")
    @ResponseStatus(HttpStatus.OK)
    @PreAuthorize("isAuthenticated()")
    @ApiPaginationParams
    @ApiOperation("Get all tags defined for sequences")
    public ResponsePage<String> getPageWithTags(
            PaginationParam paginationParam,
            SortParam sortParam
    ) {
        Pageable pageable = paginationParam.toPageable(sortParam.toSort());
        Page<String> sequencePage = sequenceService.findAllTags(pageable);
        return ResponsePage.responsePage(sequencePage);
    }

    @PatchMapping("/{id}/nucleic-counts")
    @ResponseStatus(HttpStatus.OK)
    @PreAuthorize("isAuthenticated()")
    @ApiOperation("Recount and get nucleic counts statistic")
    public Response<Sequence> patchNucleicCounts(
            @PathVariable UUID id
    ) {
        Sequence sequence = sequenceService.updateNucleicCounts(id);
        return Response.response(sequence);
    }


    private Sequence sequenceCreator(ImportSequenceFrom request,
                                     Authentication authentication) {
        return toSequence(
                request.getName(),
                request.getType(),
                request.getCircular(),
                request.getTags(),
                JwtAuthenticationUtils.toUser(authentication));
    }

    @RequestMapping(
            path = "/import/file",
            method = RequestMethod.POST,
            consumes = MediaType.MULTIPART_FORM_DATA_VALUE,
            produces = MediaType.APPLICATION_JSON_VALUE)
    @ResponseStatus(HttpStatus.CREATED)
    @PreAuthorize("isAuthenticated()")
    @ApiOperation("Create sequence from file. (null char is forbiden in fasta comments)")
    public Response<Sequence> importSequenceFromFile(
            @ApiIgnore Authentication authentication,
            @RequestPart String json,
            @RequestPart MultipartFile file
    ) {
        ImportSequenceFromFileRequest request = jsonService.fromJson(json, ImportSequenceFromFileRequest.class);
        try {
            Sequence initial = sequenceCreator(request, authentication);
            return Response.response(importSequenceByFormat(initial, request.getFormat(), file));
        } catch (UnsupportedOperationException e) {
            throw new InternalServerException("Get data from MultipartFile failed.", e);
        }
    }

    @PostMapping("import/text")
    @ResponseStatus(HttpStatus.CREATED)
    @PreAuthorize("isAuthenticated()")
    @ApiOperation("Create sequence from plain text.(null char is forbiden in fasta comments)")
    public Response<Sequence> importSequenceFromText(
            @ApiIgnore Authentication authentication,
            @RequestBody @Valid ImportSequenceFromTextRequest request
    ) {
        Sequence initial = sequenceCreator(request, authentication);
        return Response.response(importSequenceByFormat(initial, request.getFormat(), request.getData().getBytes(StandardCharsets.UTF_8)));
    }

    /**
     * Import from NCBI Requests object definition.
     */
    @Data
    static class ImportNCBIRequest {

        @Data
        static class NCBI {

            public String getNcbiId() {
                return ncbiId.trim();
            }

            @NotNull
            private String ncbiId;

            private String name;

            private Boolean circular;

            private SequenceType type;

            private Set<String> tags;
        }

        private List<NCBI> ncbis;

        private Boolean circular;

        private SequenceType type;

        private Set<String> tags;
    }

    @PostMapping("/import/ncbi")
    @ResponseStatus(HttpStatus.CREATED)
    @PreAuthorize("isAuthenticated()")
    @ApiOperation("Create sequence from ncbi.")
    public ResponseList<Sequence> importSequenceFromNCBI(
            @ApiIgnore Authentication authentication,
            @RequestBody @Valid ImportNCBIRequest request
    ) {
        List<Sequence> sequences = request.getNcbis()
                .stream()
                .map(ncbi ->
                        sequenceService.importNCBI(
                                toSequence(
                                        ncbi.getName(),
                                        (ncbi.getType() == null) ? request.getType() : ncbi.getType(),
                                        (ncbi.getCircular() == null) ? request.getCircular() : ncbi.getCircular(),
                                        TagUtils.union(request.getTags(), ncbi.getTags()),
                                        JwtAuthenticationUtils.toUser(authentication)),
                                ncbi.getNcbiId())
                ).collect(Collectors.toList());

        return ResponseList.responseList(sequences);
    }

    @GetMapping("/{id}")
    @ResponseStatus(HttpStatus.OK)
    @PreAuthorize("isAuthenticated()")
    @ApiOperation("Get one sequence by ID")
    public Response<Sequence> getSequenceByID(
            @PathVariable UUID id
    ) {
        Sequence sequence = sequenceService.findOne(id);
        return Response.response(sequence);
    }

    @PutMapping("/{id}/tags")
    @ResponseStatus(HttpStatus.ACCEPTED)
    @PreAuthorize("isAuthenticated()")
    @ApiOperation("Modify tags")
    public Response<Sequence> modifyTags(
            @PathVariable UUID id,
            @RequestBody TagModifyRequest request) {
        Sequence sequence = sequenceService.modifyTags(id, request.getTags());
        return Response.response(sequence);
    }

    @DeleteMapping("/{id}")
    @ResponseStatus(HttpStatus.NO_CONTENT)
    @PreAuthorize("isAuthenticated()")
    @ApiOperation("Delete one sequence by ID")
    public void deleteSequence(
            @PathVariable UUID id) {
        sequenceService.delete(id);
    }


    private static Sequence toSequence(String name, SequenceType type, Boolean circular, Set<String> tags, User owner) {
        return Sequence.builder()
                .name(name)
                .type(type)
                .circular(circular)
                .tags(tags)
                .owner(owner)
                .build();
    }

    private Sequence importSequenceByFormat(Sequence sequence, ImportSequenceFormat format, byte[] input) {
        switch (format) {
            case PLAIN:
                return sequenceService.importPlainFormatFromText(sequence, input);
            case FASTA:
                return sequenceService.importFASTAFormatFromText(sequence, input);
            default: // If no origin type is define, use plain type import
                return sequenceService.importPlainFormatFromText(sequence, input);
        }
    }

    private Sequence importSequenceByFormat(Sequence sequence, ImportSequenceFormat format, MultipartFile input) {
        switch (format) {
            case PLAIN:
                return sequenceService.importPlainFormatFromFile(sequence, input);
            case FASTA:
                return sequenceService.importFASTAFormatFromFile(sequence, input);
            default: // If no origin type is define, use plain type import
                return sequenceService.importPlainFormatFromFile(sequence, input);
        }
    }

}
