package cz.mendelu.dnaAnalyser.sequence.data;

import cz.mendelu.dnaAnalyser.sequence.Sequence;
import cz.mendelu.dnaAnalyser.sequence.inport.ImportSequenceService;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.io.LineIterator;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Component;

import java.io.*;
import java.nio.Buffer;
import java.nio.ByteBuffer;
import java.nio.channels.FileChannel;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.nio.file.StandardOpenOption;
import java.util.Iterator;
import java.util.Objects;
import java.util.UUID;
import java.util.stream.Stream;

import static java.text.MessageFormat.format;

@Slf4j
@Component
public class SequenceDataRepository {
    private static final String FASTA = "fasta";
    private static final String INTERNAL_FORMAT = "sequence-buffer";
    private static final int SHARED_CPU_CAHCE = 3000000;//it is ideal that data fit into cache... threads may starve for data (wait for ram)
    private static final int MAX_LINE_LENGTH = SHARED_CPU_CAHCE / Runtime.getRuntime().availableProcessors();

    private Path storageDir;

    @Value("${sequence.dir}")
    public void setStorageDir(String storageDir) {
        this.storageDir = Paths.get(storageDir);
        // Create if not exist
        if (Files.notExists(this.storageDir)) {
            try {
                Files.createDirectories(this.storageDir);
            } catch (IOException e) {
                throw new UnsupportedOperationException(format("Create sequence {0} dir failed.", storageDir), e);
            }
        }
    }

    public Path saveRawBuffer(UUID bufferId, String format, byte[] buffer) {
        Path file = getPath(bufferId, format);
        assert Files.notExists(file) : "Sequence buffer data-file can't by overridden.";
        try {
            Files.write(file, buffer);
            return file;
        } catch (IOException e) {
            throw new UnsupportedOperationException(e);
        }
    }

    // precessing to make  lines long enough to be worth processing in parallel
    // but not that long that cpus would be running out of cache constantly
    public Path saveRawBufferFromFile(UUID bufferId, String format, InputStream buffer) {
        Path path = getPath(bufferId, format);
        assert Files.notExists(path) : "Sequence buffer data-file can't by overridden.";
        //light pre processing
        try (BufferedWriter writer = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(path.toFile()), StandardCharsets.UTF_8))) {
            try (LineIterator raw = new LineIterator(new InputStreamReader(buffer))) {
                String line;
                long currentLength = 0;
                if (format.equals(FASTA)) {
                    boolean previsousWasFasta = false;
                    while (raw.hasNext()) {
                        line = raw.next();
                        //found fasta comment that must be on separate line... reset counter put it on separate line
                        if (line.startsWith(">") || line.startsWith(";")) {
                            currentLength = 0;
                            if (!previsousWasFasta) {
                                writer.write("\n");
                                previsousWasFasta = true;
                            } else {
                                //agregating fasta lines to make extraction stateless
                                // look here
                                //https://docs.oracle.com/javase/tutorial/collections/streams/parallelism.html#stateful_lambda_expressions
                                writer.write(ImportSequenceService.FASTA_PSEOUDO_LINE);

                            }

                            writer.write(line);
                        } else {
                            if (previsousWasFasta) {
                                previsousWasFasta = false;
                                writer.write("\n");
                            }
                            currentLength += line.length();
                            if (currentLength >= MAX_LINE_LENGTH) {
                                writer.write("\n");
                                currentLength = 0;
                            }
                            writer.write(line);
                        }
                    }
                } else {
                    while (raw.hasNext()) {
                        line = raw.next();
                        currentLength += line.length();
                        if (currentLength >= MAX_LINE_LENGTH) {
                            writer.write("\n");
                            currentLength = 0;
                        }
                        writer.write(line);
                    }
                }
                return path;
            }
        } catch (IOException e) {
            try {
                Files.delete(path);
            } catch (IOException ex) {
                ex = new IOException("during emergency deletion(during exception) of partially written uploaded file " +
                        "exception encountered", ex);
                e.addSuppressed(ex);
            }
            throw new UnsupportedOperationException(e);
        } finally {
            try {
                buffer.close();
            } catch (IOException e) {
                log.warn("can't close stream of uploaded data.", e);
            }
        }
    }

    public Stream<String> loadRawBufferToStream(Sequence sequence, String format) {
        try {
            Path file = getPath(sequence.getBufferId(), format);
            return Files.lines(file);
        } catch (IOException e) {
            throw new UnsupportedOperationException(e);
        }
    }

    public void deleteRawBuffer(UUID bufferId, String format) {
        Path file = getPath(bufferId, format);
        try {
            Files.delete(file);
        } catch (IOException e) {
            log.warn(format("Delete sequence data with format: {0} failed.", format), e);
        }
    }

    public Path save(UUID bufferId, Iterator<ByteBuffer> buffers) {
        Path file = getPath(bufferId, INTERNAL_FORMAT);
        assert Files.notExists(file) : "Sequence buffer data-file can't by overridden.";
        try (FileChannel output = new FileOutputStream(file.toFile()).getChannel()) {
            ByteBuffer current;
            while (buffers.hasNext()) {
                current = buffers.next().asReadOnlyBuffer();//making sure we have independent counters on buffer
                ((Buffer)current).position(0); // Fix Java bug
                output.write(current);
            }
            //try with resources closes file itself
            return file;
        } catch (IOException e) {
            try {
                //delete unfinished file?
                Files.delete(file);
            } catch (IOException deleteFail) {
                e.addSuppressed(new IOException("Exception In exception during cleanup of incompletely written file", deleteFail));
            }
            throw new UnsupportedOperationException("Exception during saving file", e);
        }
    }

    public SequenceData load(Sequence sequence) {
        Path file = getPath(sequence.getBufferId(), INTERNAL_FORMAT);
        final long fileSize;
        try {
            fileSize = Files.size(file);
            assert fileSize < Integer.MAX_VALUE : "Datafile is to big.";
            return SequenceData.wrap(sequence, FileChannel.open(file, StandardOpenOption.READ).map(FileChannel.MapMode.READ_ONLY, 0, fileSize));
        } catch (IOException e) {
            throw new UnsupportedOperationException(e);
        }
    }

    public void delete(UUID bufferId) {
        Path file = getPath(bufferId, INTERNAL_FORMAT);
        try {
            Files.delete(file);
        } catch (IOException e) {
            log.warn("Delete sequence data failed.", e);
        }
    }

    private Path getPath(UUID bufferId, String format) {
        String suffix = (Objects.equals(format, INTERNAL_FORMAT))
                ? ""
                : "." + format.toLowerCase();
        return storageDir.resolve(bufferId + suffix);
    }
}
